const adminCtrl = require("../controllers/admin.controller");
const config = require("../config/common.config");
const express = require("express");
const multer = require("multer");
const fs = require("fs");
const router = express.Router();
const ResponseFormatter = require("./../utils/response-formatter");
const formatter = new ResponseFormatter();
const adminValidation = require("../validations/adminValidations");
const routeMiddlewares = require("./routeMiddlewares");
const publicPath = global.basedir + "/public/";

const multerStorage = multer.diskStorage({
    destination: (req, file, cb) => {
        if (file.fieldname == "category_image") {
            if (!fs.existsSync("public/category")) {
                fs.mkdirSync("public/category");
            }
            cb(null, "public/category");
        }
        if (file.fieldname == "pet_image") {
            if (!fs.existsSync("public/petType")) {
                fs.mkdirSync("public/petType");
            }
            cb(null, "public/petType");
        }
        if (file.fieldname == "product_image") {
            if (!fs.existsSync("public/product")) {
                fs.mkdirSync("public/product");
            }
            cb(null, "public/product");
        }
        if (file.fieldname == "profile") {
            if (!fs.existsSync("public/profile")) {
                fs.mkdirSync("public/profile");
            }
            cb(null, "public/profile");
        }

    },
    filename: (req, file, cb) => {
        const ext = file.mimetype.split("/")[1];
        cb(null, `img_${Date.now()}.${ext}`);
    },
});

const upload = multer({ storage: multerStorage });

router.post("/login", routeMiddlewares.validateRequest(adminValidation.login), (req, res) => {
    adminCtrl.login(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['login_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post('/forgotPassword', routeMiddlewares.validateRequest(adminValidation.forgotPasswordSchema), (req, res, next) => {
    adminCtrl.forgotPassword(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['forgot_password_successfully'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
})

router.post('/verifyOTP', routeMiddlewares.validateRequest(adminValidation.verifyOTPSchema), (req, res, next) => {
    adminCtrl.verifyOTP(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['otp_verification_successfully'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
})

router.post('/resetPassword', routeMiddlewares.validateRequest(adminValidation.resetPasswordSchema), (req, res, next) => {
    adminCtrl.resetPassword(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['password_change_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
})

router.post("/userList", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.userSchema), (req, res) => {
    adminCtrl.userList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/category", upload.single('category_image'), routeMiddlewares.validateRequest(adminValidation.categorySchema), routeMiddlewares.authAdmin, (req, res) => {
    adminCtrl.category(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['category_added'], true))
        })
        .catch((err) => {
            if (req.file) {
                fs.unlinkSync(publicPath + "category/" + req.file.filename)
            }
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/categoryList", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.categoryListSchema), (req, res) => {
    adminCtrl.categoryList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/categoryUpdate", upload.single('category_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.categoryUpdateSchema), (req, res) => {
    adminCtrl.categoryUpdate(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['updated'], true))
        })
        .catch((err) => {
            if (req.file) {
                fs.unlinkSync(publicPath + "category/" + req.file.filename)
            }
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/categoryDelete", upload.single('category_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.categoryDeleteSchema), (req, res) => {
    adminCtrl.categoryDelete(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['deleted'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/product", upload.single('product_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.productSchema), (req, res) => {
    adminCtrl.product(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['product_added'], true))
        })
        .catch((err) => {
            if (req.file && req.file.profile) {
                fs.unlinkSync(publicPath + "product/" + req.file.filename);
            }
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/productUpdate", upload.single('product_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.productUpdateSchema), (req, res) => {
    adminCtrl.productUpdate(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['updated'], true))
        })
        .catch((err) => {
            if (req.file && req.file.profile) {
                fs.unlinkSync(publicPath + "product/" + req.file.filename);
            }
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/productDelete", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.productDeleteSchema), (req, res) => {
    adminCtrl.productDelete(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['deleted'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/productList", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.productListSchema), (req, res) => {
    adminCtrl.productList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/createPetType", upload.single('pet_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.createPetTypeSchema), (req, res) => {
    adminCtrl.createPetType(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['pet_created'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/createPetBreedSpecies", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.createPetBreedSpeciesSchema), (req, res) => {
    adminCtrl.createPetBreedSpecies(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['pet_created'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/updatePetType", upload.single('pet_image'), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.updatePetTypeSchema), (req, res) => {
    adminCtrl.updatePetType(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['pet_updated'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/deletePetType", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.deletePetTypeSchema), (req, res) => {
    adminCtrl.deletePetType(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['pet_deleted'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/PetTypeList", routeMiddlewares.authAdmin, (req, res) => {
    adminCtrl.PetTypeList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/prescriptionList", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.prececriptionListSchema), (req, res) => {
    adminCtrl.prescriptionList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/editDoctorInfo", upload.single("profile"), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.editDoctorInfoSchema), (req, res) => {
    adminCtrl.editDoctorInfo(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['doctor_info_update'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/editUserInfo", upload.single("profile"), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.editUserInfoSchema), (req, res) => {
    adminCtrl.editUserInfo(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['user_info_update'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/deleteDoctor", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.deleteDoctorSchema), (req, res) => {
    adminCtrl.deleteDoctor(req)
        .then((result) => {
            res.send(formatter.formatResponse({}, 1, config.messages[result], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/hideUnHideDoctor", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.hideUnHideDoctor), (req, res) => {
    adminCtrl.hideUnHideDoctor(req)
        .then((result) => {
            res.send(formatter.formatResponse({}, 1, config.messages[result], true))
        }).catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/suspendUnsuspendDoctor", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.suspendUnsuspendDoctor), (req, res) => {
    adminCtrl.suspendDoctor(req)
        .then((result) => {



            res.send(formatter.formatResponse({}, 1, config.messages[result], true))
        }).catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/getPetDetail", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.petDetailSchema), (req, res) => {
    adminCtrl.getPetDetail(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages["list_success"], true))
        }).catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/updateOrderStatus", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.updateOrderSchema), (req, res) => {
    adminCtrl.updateOrderStatus(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages["order_status_updated"], true))
        }).catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/getOrderDetail", upload.none(), routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.orderDetailSchema), (req, res) => {
    adminCtrl.getOrderDetail(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages["order_detail_success"], true))
        }).catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});


router.post("/orderList", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.orderListSchema), (req, res) => {
    adminCtrl.orderList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

router.post("/getpayoutList", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.getpayoutListSchema), (req, res) => {
    adminCtrl.getpayoutList(req)
        .then((result) => {
            res.send(formatter.formatResponse(result, 1, config.messages['list_success'], true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});
router.post("/payoutStatusUpdate", routeMiddlewares.authAdmin, routeMiddlewares.validateRequest(adminValidation.payoutStatusUpdateSchema), (req, res) => {
    adminCtrl.payoutStatusUpdate(req)
        .then((result) => {
            let resMsg = config.messages['payout_req_accept']
            if (req.body.status == "reject") {
                resMsg = config.messages['payout_req_reject']
            }
            res.send(formatter.formatResponse(result, 1, resMsg, true))
        })
        .catch((err) => {
            res.send(formatter.formatResponse({}, 0, err, false))
        })
});

module.exports = router