const userCtrl = require("../controllers/user.controller");
const userHomeCtrl = require("../controllers/userhome.controller");
const config = require("../config/common.config");
const express = require("express");
const ResponseFormatter = require("../utils/response-formatter");
const userValidation = require("../validations/userValidations");
const routeMiddlewares = require("../routes/routeMiddlewares");
const multer = require("multer");
const fs = require("fs");
const publicPath = basedir + "/public/";

const multerStorage = multer.diskStorage({
    destination: (req, file, cb) => {
        if (file.fieldname == "document") {
            if (!fs.existsSync("public/document")) {
                fs.mkdirSync("public/document");
            }
            cb(null, "public/document");
        }
        else if (file.fieldname == "product_image") {
            if (!fs.existsSync("public/product")) {
                fs.mkdirSync("public/product");
            }
            cb(null, "public/product");
        }
        else if (file.fieldname == "pet_profile") {
            if (!fs.existsSync("public/pet")) {
                fs.mkdirSync("public/pet");
            }
            cb(null, "public/pet");
        }
        else if (file.fieldname == "profile") {
            if (!fs.existsSync("public/profile")) {
                fs.mkdirSync("public/profile");
            }
            cb(null, "public/profile");
        }
        else {
            if (!fs.existsSync("public/pet")) {
                fs.mkdirSync("public/pet");
            }
            cb(null, "public/pet");
        }
    },
    filename: (req, file, cb) => {
        const ext = file.mimetype.split("/")[1];
        const unique = Math.floor(Math.random() * (9999 - 1000 + 1)) + 1000
        cb(null, `img_${Date.now()}${unique}.${ext}`);
    },
});

const upload = multer({ storage: multerStorage });

const formatter = new ResponseFormatter();
const router = express.Router();


router.post(
    "/refreshToken",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.userIdSchema),
    userCtrl.refreshToken
);


router.post("/signUp",
    upload.single("profile"),
    routeMiddlewares.validateRequest(userValidation.signupSchema),
    userCtrl.signUp,
);


router.post("/sendOtp",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.sendOtpSchema),
    userCtrl.sendOtp,
);


router.post(
    "/verifyOTP",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.verifyOTPSchema),
    userCtrl.verifyOTP
);


router.post(
    "/updateProfile",
    upload.single("profile"),
    routeMiddlewares.validateRequest(userValidation.updateProfileSchema),
    routeMiddlewares.authorize,
    userCtrl.updateProfile,
);


router.post(
    "/logOut",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.logoutSchema),
    routeMiddlewares.authorize,
    userCtrl.logOut
);


router.post(
    "/deleteAccount",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.deleteAccountSchema),
    routeMiddlewares.authorize,
    userCtrl.deleteAccount
);


router.post(
    "/contactUs",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.contactUsSchema),
    routeMiddlewares.authorize,
    userCtrl.contactUs
);

router.post(
    "/getProfile",
    upload.none(),
    routeMiddlewares.authorize,
    userCtrl.getProfile
);

router.post(
    "/address",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.addressSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.address
);

router.post(
    "/updateAddress",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.updateAddressSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.updateAddress
);

router.post(
    "/deleteAddress",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.deleteAddressSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.deleteAddress
);

router.get(
    "/AddressList",
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.AddressList
);

router.post(
    "/categoryList",
    upload.none(),
    routeMiddlewares.checkVersion,
    routeMiddlewares.authorize,
    routeMiddlewares.validateRequest(userValidation.catTypeSchema),
    userHomeCtrl.categoryList
);

router.post(
    "/productDetail",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.productDeatilSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.productDetail
);

router.post(
    "/productList",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.productListSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.productList
);

router.post(
    "/addToCart",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.addToCartSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.addToCart
);

router.get(
    "/cartDetail",
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.cartDetail
);

router.post(
    "/placeOrder",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.placeOrderSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.placeOrder
);

router.post(
    "/updateCart",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.updateCartSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.updateCart
);

router.post(
    "/petProfile",
    upload.any(),
    routeMiddlewares.validateRequest(userValidation.petProfileSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.petProfile
);

router.get(
    "/petProfileList",
    upload.any(),
    routeMiddlewares.authorize,
    userHomeCtrl.petProfileList
);

router.post(
    "/updatePetProfile",
    upload.any(),
    routeMiddlewares.validateRequest(userValidation.updatePetProfileSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.updatePetProfile
);

router.post(
    "/deletePetProfile",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.deletePetProfileSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.deletePetProfile
);

router.get(
    "/issueList",
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.issueList
);

router.post(
    "/doctorList",
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.doctorSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.doctorList
);

router.post(
    "/consultation",
    upload.array('image'),
    routeMiddlewares.validateRequest(userValidation.consultationSchema),
    routeMiddlewares.checkVersion,
    routeMiddlewares.authorize,
    userHomeCtrl.consultation
);

router.post(
    "/priscibeList",
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.priscibeList
);

// router.post(
//     '/orderPlacePrescribe',
//     upload.none(),
//     routeMiddlewares.validateRequest(userValidation.orderPlacePrescribeSchema),
//     routeMiddlewares.authorize,
//     userHomeCtrl.orderPlacePrescribe,
// )

router.get(
    '/orderHistory',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.orderHistory,
)
router.get(
    '/paymentHistory',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.paymentHistory,
)

router.post(
    '/chatAi',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.chatAiSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.chatAi,
)
router.get(
    '/chatAiHistory',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.chatAiHistory,
)
router.post(
    '/petNutrition',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.petNutritionSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.petNutrition,
)

router.post(
    '/consultationDetail',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.consultationDetailSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.consultationDetail,
)
router.post(
    '/notificationList',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.notificationList,
)
router.post(
    '/consultationSlot',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.consultationSlotSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.consultationSlot,
)
router.post(
    '/consultationClose',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.consultationCloseSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.consultationClose,
)
router.post(
    '/refundPayment',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.refundPaymentSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.refundPayment,
)
router.post(
    '/addBreed',
    upload.none(),
    userHomeCtrl.addBreed,
)
router.post(
    '/getBreed',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.getBreed,
)
router.post(
    '/addRecipes',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.addRecipes,
)
router.post(
    '/getRecipesHistory',
    upload.none(),
    routeMiddlewares.authorize,
    userHomeCtrl.getRecipesHistory,
)
router.post(
    '/purchaseRecipePlan',
    upload.none(),
    routeMiddlewares.validateRequest(userValidation.purchaseRecipeSchema),
    routeMiddlewares.authorize,
    userHomeCtrl.purchaseRecipePlan,
)

module.exports = router;
